from typing import Any, List, Optional
from rekuest.api.schema import AssignationLogLevel
from rekuest.messages import Assignation
from rekuest.postmans.utils import ReservationContract
from fluss.api.schema import ArkitektNodeFragment
from reaktion.atoms.generic import MapAtom
from reaktion.events import Returns


class TemplateMapAtom(MapAtom):
    node: ArkitektNodeFragment
    contract: ReservationContract

    async def map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}
        returns = await self.contract.aassign(
            *args,
            **defaults,
            alog=self.alog_arkitekt,
            parent=self.assignation.assignation
        )
        return returns
        # return await self.contract.aassign(*args)

    async def alog_arkitekt(
        self, assignation: Assignation, level: AssignationLogLevel, message: str
    ):
        if self.alog:
            await self.alog(self.node.id, level, message)



import asyncio
from reaktion.agent import ReaktionAgent

from rekuest.contrib.fakts.websocket_agent_transport import FaktsWebsocketAgentTransport
from rekuest.api.schema import (
    PortInput,
    DefinitionInput,
    TemplateFragment,
    acreate_template,
    adelete_node,
    afind,
)
from fakts.fakts import Fakts
from fluss.api.schema import (
    FlowFragment,
    LocalNodeFragment,
)
from rekuest.widgets import SliderWidget, StringWidget
from arkitekt.apps.fluss import FlussApp
from arkitekt.apps.unlok import UnlokApp
from arkitekt.apps.rekuest import ArkitektRekuest, RekuestApp
from reaktion.utils import infer_kind_from_graph
from fakts.grants.remote.device_code import DeviceCodeGrant
from fakts.fakts import Fakts
from fakts.grants import CacheGrant
from herre.grants import CacheGrant as HerreCacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.grants.fakts import FaktsGrant
from fakts.grants.remote import Manifest
from fakts.discovery import StaticDiscovery
from herre import Herre


class ConnectedApp(FlussApp, RekuestApp, UnlokApp):
    pass


identifier = "fluss"
version = "latest"
url = "http://localhost:8000/f/"

app = ConnectedApp(
    fakts=Fakts(
        grant=CacheGrant(
            cache_file=f"{identifier}-{version}_cache.json",
            hash=f"{identifier}-{version}-{url}",
            grant=DeviceCodeGrant(
                manifest=Manifest(
                    identifier=identifier,
                    version=version,
                    scopes=["read", "write"],
                ),
                discovery=StaticDiscovery(base_url=url),
            ),
        )
    ),
    herre=Herre(
        grant=HerreCacheGrant(
            cache_file=f"{identifier}-{version}_herre_cache.json",
            hash=f"{identifier}-{version}-{url}",
            grant=RefreshGrant(grant=FaktsGrant()),
        ),
    ),
    rekuest=ArkitektRekuest(
        agent=ReaktionAgent(
            transport=FaktsWebsocketAgentTransport(fakts_group="rekuest.agent")
        )
    ),
)


@app.rekuest.register(
    widgets={"description": StringWidget(as_paragraph=True)},
    interfaces=["fluss:deploy"],
)
async def deploy_graph(
    flow: FlowFragment,
    name: str = None,
    description: str = None,
) -> TemplateFragment:
    """Deploy Flow

    Deploys a Flow as a Template

    Args:
        graph (FlowFragment): The Flow
        name (str, optional): The name of this Incarnation
        description (str, optional): The name of this Incarnation

    Returns:
        TemplateFragment: The created template
    """
    assert flow.name, "Graph must have a Name in order to be deployed"

    print([x.dict(by_alias=True) for x in flow.graph.args])
    print([x.dict(by_alias=True) for x in flow.graph.returns])

    # assert localnodes are in the definitionregistry
    localNodes = [x for x in flow.graph.nodes if isinstance(x, LocalNodeFragment)]
    for node in localNodes:
        assert node.hash, f"LocalNode {node.name} must have a definition"
        assert (
            node.hash in app.rekuest.agent.nodeHashActorMap
        ), f"LocalNode {node.name} is not registered with the agent of this instance"

    args = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.args]
    returns = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.returns]

    template = await acreate_template(
        DefinitionInput(
            name=name or flow.workspace.name,
            interface=f"flow-{flow.id}",
            kind=infer_kind_from_graph(flow.graph),
            args=args,
            returns=returns,
            description=description,
            interfaces=["workflow", f"diagram:{flow.workspace.id}", f"flow:{flow.id}"],
        ),
        instance_id=app.rekuest.agent.instance_id,
        params={"flow": flow.id},
        extensions=["flow"],
    )

    return template


@app.rekuest.register()
async def undeploy_graph(
    flow: FlowFragment,
):
    """Undeploy Flow

    Undeploys graph, no user will be able to reserve this graph anymore

    Args:
        graph (FlowFragment): The Flow

    """
    assert flow.name, "Graph must have a Name in order to be deployed"

    x = await afind(interface=flow.hash)

    await adelete_node(x)
    return None


@app.rekuest.register(widgets={"interval": SliderWidget(min=0, max=100)})
async def timer(interval: int = 4) -> int:
    """Timer

    A simple timer that prints the current time every interval seconds

    Args:
        interval (int, optional): The interval in seconds. Defaults to 4.

    Returns:
        int: The current interval (iteration)

    """
    i = 0
    while True:
        i += 1
        yield i
        await asyncio.sleep(interval)


async def main():
    async with app:
        await app.rekuest.run()


if __name__ == "__main__":
    asyncio.run(main())



from reaktion.actor import FlowActor
from rekuest.agents.errors import ProvisionException
from rekuest.agents.stateful import StatefulAgent
import logging
from rekuest.actors.base import Actor
from fluss.api.schema import aget_flow
from rekuest.api.schema import aget_template, NodeKind
from rekuest.messages import Provision

from rekuest.contrib.fakts.websocket_agent_transport import FaktsWebsocketAgentTransport
from rekuest.api.schema import (
    PortInput,
    DefinitionInput,
    TemplateFragment,
    acreate_template,
    adelete_node,
    afind,
)
from fakts.fakts import Fakts
from fluss.api.schema import (
    FlowFragment,
    LocalNodeFragment,
    GraphNodeFragment,
)
from reaktion.utils import infer_kind_from_graph
from rekuest.widgets import SliderWidget, StringWidget

logger = logging.getLogger(__name__)


class ReaktionAgent(StatefulAgent):
    async def aspawn_actor(self, prov: Provision) -> Actor:
        """Spawns an Actor from a Provision"""
        try:
            actor_builder = self._templateActorBuilderMap[prov.template]
            actor = actor_builder(provision=prov, transport=self.transport)
        except KeyError:
            try:
                x = await aget_template(prov.template)
                assert "flow" in x.params, "Template is not a flow"

                t = await aget_flow(id=x.params["flow"])
                actor = FlowActor(
                    provision=prov,
                    transport=self.transport,
                    agent=self,
                    flow=t,
                    is_generator=x.node.kind == NodeKind.GENERATOR,
                )

            except Exception as e:
                raise ProvisionException("No Actor Builders found for template") from e

        await actor.arun()
        self.provisionActorMap[prov.provision] = actor
        return actor

    async def aregister_definitions(self):
        self.definition_registry.register(
            self.deploy_graph,
            structure_registry=self.definition_registry.structure_registry,
            widgets={"description": StringWidget(as_paragraph=True)},
            interfaces=["fluss:deploy"],
        )
        self.definition_registry.register(
            self.undeploy_graph,
            structure_registry=self.definition_registry.structure_registry,
            interfaces=["fluss:undeploy"],
        )

        return await super().aregister_definitions()

    async def deploy_graph(
        self,
        flow: FlowFragment,
        name: str = None,
        description: str = None,
    ) -> TemplateFragment:
        """Deploy Flow

        Deploys a Flow as a Template

        Args:
            graph (FlowFragment): The Flow
            name (str, optional): The name of this Incarnation
            description (str, optional): The name of this Incarnation

        Returns:
            TemplateFragment: The created template
        """
        assert flow.name, "Graph must have a Name in order to be deployed"

        print([x.dict(by_alias=True) for x in flow.graph.args])
        print([x.dict(by_alias=True) for x in flow.graph.returns])

        # assert localnodes are in the definitionregistry
        localNodes = [x for x in flow.graph.nodes if isinstance(x, LocalNodeFragment)]
        graphNodes = [x for x in flow.graph.nodes if isinstance(x, GraphNodeFragment)]
        assert len(graphNodes) == 0, "GraphNodes are not supported yet"

        for node in localNodes:
            assert node.hash, f"LocalNode {node.name} must have a definition"
            assert (
                node.hash in self.nodeHashActorMap
            ), f"LocalNode {node.name} is not registered with the agent of this instance"

        args = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.args]
        returns = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.returns]

        template = await acreate_template(
            DefinitionInput(
                name=name or flow.workspace.name,
                interface=f"flow-{flow.id}",
                kind=infer_kind_from_graph(flow.graph),
                args=args,
                returns=returns,
                description=description,
                interfaces=[
                    "workflow",
                    f"diagram:{flow.workspace.id}",
                    f"flow:{flow.id}",
                ],
            ),
            instance_id=self.instance_id,
            params={"flow": flow.id},
            extensions=["flow"],
        )

        return template

    async def undeploy_graph(
        flow: FlowFragment,
    ):
        """Undeploy Flow

        Undeploys graph, no user will be able to reserve this graph anymore

        Args:
            graph (FlowFragment): The Flow

        """
        assert flow.name, "Graph must have a Name in order to be deployed"

        x = await afind(interface=flow.hash)

        await adelete_node(x)
        return None



from typing import Any, List, Optional
from rekuest.postmans.utils import RPCContract
from fluss.api.schema import LocalNodeFragment
from reaktion.atoms.generic import MapAtom, MergeMapAtom
from reaktion.events import Returns
import logging

logger = logging.getLogger(__name__)


class LocalMapAtom(MapAtom):
    node: LocalNodeFragment
    contract: RPCContract

    async def map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        returns = await self.contract.aassign(
            **defaults, parent=self.assignation.assignation
        )
        return returns
        # return await self.contract.aassign(*args)


class LocalMergeMapAtom(MergeMapAtom):
    node: LocalNodeFragment
    contract: RPCContract

    async def merge_map(self, args: Returns) -> Optional[List[Any]]:
        defaults = self.node.defaults or {}

        stream_one = self.node.instream[0]
        for arg, item in zip(args, stream_one):
            defaults[item.key] = arg

        async for r in self.contract.astream(
            **defaults, parent=self.assignation.assignation
        ):
            yield r
